/*
 *
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */

package org.apache.metron.profiler;

import java.util.concurrent.TimeUnit;

/**
 * The Profiler captures a ProfileMeasurement once every ProfilePeriod.  There can be
 * multiple ProfilePeriods every hour.
 */
public class ProfilePeriod {

  /**
   * A monotonically increasing number identifying the period.  The first period is 0
   * and began at the epoch.
   */
  private long period;

  /**
   * The duration of each period in milliseconds.
   */
  private long durationMillis;

  /**
   * @param epochMillis A timestamp contained somewhere within the profile period.
   * @param duration The duration of each profile period.
   * @param units The units of the duration; hours, minutes, etc.
   */
  public ProfilePeriod(long epochMillis, long duration, TimeUnit units) {
    this.durationMillis = units.toMillis(duration);
    this.period = epochMillis / durationMillis;
  }

  /**
   * When this period started in milliseconds since the epoch.
   */
  public long getStartTimeMillis() {
    return period * durationMillis;
  }

  /**
   * Returns the next ProfilePeriod in time.
   */
  public ProfilePeriod next() {
    long nextStart = getStartTimeMillis() + durationMillis;
    return new ProfilePeriod(nextStart, durationMillis, TimeUnit.MILLISECONDS);
  }

  public long getPeriod() {
    return period;
  }

  public long getDurationMillis() {
    return durationMillis;
  }

  @Override
  public boolean equals(Object o) {
    if (this == o) return true;
    if (o == null || getClass() != o.getClass()) return false;

    ProfilePeriod that = (ProfilePeriod) o;
    if (period != that.period) return false;
    return durationMillis == that.durationMillis;
  }

  @Override
  public int hashCode() {
    int result = (int) (period ^ (period >>> 32));
    result = 31 * result + (int) (durationMillis ^ (durationMillis >>> 32));
    return result;
  }

  @Override
  public String toString() {
    return "ProfilePeriod{" +
            "period=" + period +
            ", durationMillis=" + durationMillis +
            '}';
  }
}
